#
# Revelation 0.1.1 - a password manager for GNOME 2
# http://oss.wired-networks.net/revelation/
#
# Module containing data-related functionality
#
#
# Copyright (c) 2003-2004 Erik Grinaker
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
#


import gtk, revelation, gobject, libxml2, string
from Crypto.Cipher import AES

ENTRYSTORE_COL_NAME = 0
ENTRYSTORE_COL_ICON = 1
ENTRYSTORE_COL_TYPE = 2
ENTRYSTORE_COL_DESC = 3
ENTRYSTORE_COL_UPDATED = 4
ENTRYSTORE_COL_FIELDS = 5


ENTRY_FOLDER			= "folder"
ENTRY_ACCOUNT_CREDITCARD	= "creditcard"
ENTRY_ACCOUNT_DATABASE		= "database"
ENTRY_ACCOUNT_DOOR		= "door"
ENTRY_ACCOUNT_EMAIL		= "email"
ENTRY_ACCOUNT_FTP		= "ftp"
ENTRY_ACCOUNT_GENERIC		= "generic"
ENTRY_ACCOUNT_PHONE		= "phone"
ENTRY_ACCOUNT_SHELL		= "shell"
ENTRY_ACCOUNT_USENET		= "usenet"
ENTRY_ACCOUNT_WEBSITE		= "website"


FIELD_GENERIC_CODE		= "generic-code"
FIELD_GENERIC_EMAIL		= "generic-email"
FIELD_GENERIC_HOSTNAME		= "generic-hostname"
FIELD_GENERIC_LOCATION		= "generic-location"
FIELD_GENERIC_PASSWORD		= "generic-password"
FIELD_GENERIC_PIN		= "generic-pin"
FIELD_GENERIC_PORT		= "generic-port"
FIELD_GENERIC_URL		= "generic-url"
FIELD_GENERIC_USERNAME		= "generic-username"

FIELD_CREDITCARD_CARDTYPE	= "creditcard-cardtype"
FIELD_CREDITCARD_CARDNUMBER	= "creditcard-cardnumber"
FIELD_CREDITCARD_CCV		= "creditcard-ccv"
FIELD_CREDITCARD_EXPIRYDATE	= "creditcard-expirydate"

FIELD_PHONE_PHONENUMBER		= "phone-phonenumber"


FIELD_TYPE_EMAIL		= "email"
FIELD_TYPE_PASSWORD		= "password"
FIELD_TYPE_TEXT			= "text"
FIELD_TYPE_URL			= "url"

ENCRYPT_KEYSIZE			= 32

UNDO				= "undo"
REDO				= "redo"


class EntryTypes(object):

	def __init__(self):

		self.entrytype = {

			ENTRY_FOLDER			: {
				"name"		: "Folder",
				"icon"		: revelation.ui.STOCK_FOLDER,
				"fields"	: []
			},

			ENTRY_ACCOUNT_CREDITCARD	: {
				"name"		: "Creditcard",
				"icon"		: revelation.ui.STOCK_ACCOUNT_CREDITCARD,
				"fields"	: [ FIELD_CREDITCARD_CARDTYPE, FIELD_CREDITCARD_CARDNUMBER, FIELD_CREDITCARD_EXPIRYDATE, FIELD_CREDITCARD_CCV, FIELD_GENERIC_PIN ]
			},

			ENTRY_ACCOUNT_DATABASE		: {
				"name"		: "Database",
				"icon"		: revelation.ui.STOCK_ACCOUNT_DATABASE,
				"fields"	: [ FIELD_GENERIC_HOSTNAME, FIELD_GENERIC_USERNAME, FIELD_GENERIC_PASSWORD ]
			},

			ENTRY_ACCOUNT_DOOR		: {
				"name"		: "Door lock",
				"icon"		: revelation.ui.STOCK_ACCOUNT_DOOR,
				"fields"	: [ FIELD_GENERIC_LOCATION, FIELD_GENERIC_CODE ]
			},

			ENTRY_ACCOUNT_EMAIL		: {
				"name"		: "Email",
				"icon"		: revelation.ui.STOCK_ACCOUNT_EMAIL,
				"fields"	: [ FIELD_GENERIC_EMAIL, FIELD_GENERIC_HOSTNAME, FIELD_GENERIC_USERNAME, FIELD_GENERIC_PASSWORD ]
			},

			ENTRY_ACCOUNT_FTP		: {
				"name"		: "FTP",
				"icon"		: revelation.ui.STOCK_ACCOUNT_FTP,
				"fields"	: [ FIELD_GENERIC_HOSTNAME, FIELD_GENERIC_PORT, FIELD_GENERIC_USERNAME, FIELD_GENERIC_PASSWORD ]
			},

			ENTRY_ACCOUNT_GENERIC		: {
				"name"		: "Generic",
				"icon"		: revelation.ui.STOCK_ACCOUNT,
				"fields"	: [ FIELD_GENERIC_USERNAME, FIELD_GENERIC_PASSWORD ]
			},

			ENTRY_ACCOUNT_PHONE		: {
				"name"		: "Phone",
				"icon"		: revelation.ui.STOCK_ACCOUNT_PHONE,
				"fields"	: [ FIELD_PHONE_PHONENUMBER, FIELD_GENERIC_PIN ]
			},

			ENTRY_ACCOUNT_SHELL		: {
				"name"		: "Shell",
				"icon"		: revelation.ui.STOCK_ACCOUNT_SHELL,
				"fields"	: [ FIELD_GENERIC_HOSTNAME, FIELD_GENERIC_USERNAME, FIELD_GENERIC_PASSWORD ]
			},

			ENTRY_ACCOUNT_USENET		: {
				"name"		: "Usenet",
				"icon"		: revelation.ui.STOCK_ACCOUNT_USENET,
				"fields"	: [ FIELD_GENERIC_HOSTNAME, FIELD_GENERIC_USERNAME, FIELD_GENERIC_PASSWORD ]
			},

			ENTRY_ACCOUNT_WEBSITE		: {
				"name"		: "Website",
				"icon"		: revelation.ui.STOCK_ACCOUNT_WEBSITE,
				"fields"	: [ FIELD_GENERIC_URL, FIELD_GENERIC_USERNAME, FIELD_GENERIC_PASSWORD ]
			}
		}

		self.fieldtype = {

			FIELD_GENERIC_CODE		: { "name" : "Code",		"type" : FIELD_TYPE_PASSWORD },
			FIELD_GENERIC_EMAIL		: { "name" : "Email address",	"type" : FIELD_TYPE_EMAIL },
			FIELD_GENERIC_HOSTNAME		: { "name" : "Hostname",	"type" : FIELD_TYPE_TEXT },
			FIELD_GENERIC_LOCATION		: { "name" : "Location",	"type" : FIELD_TYPE_TEXT },
			FIELD_GENERIC_PASSWORD		: { "name" : "Password",	"type" : FIELD_TYPE_PASSWORD },
			FIELD_GENERIC_PIN		: { "name" : "PIN",		"type" : FIELD_TYPE_PASSWORD },
			FIELD_GENERIC_PORT		: { "name" : "Port number",	"type" : FIELD_TYPE_TEXT },
			FIELD_GENERIC_URL		: { "name" : "URL",		"type" : FIELD_TYPE_URL },
			FIELD_GENERIC_USERNAME		: { "name" : "Username",	"type" : FIELD_TYPE_TEXT },

			FIELD_CREDITCARD_CARDTYPE	: { "name" : "Card type",	"type" : FIELD_TYPE_TEXT },
			FIELD_CREDITCARD_CARDNUMBER 	: { "name" : "Card number",	"type" : FIELD_TYPE_TEXT },
			FIELD_CREDITCARD_CCV		: { "name" : "CCV number",	"type" : FIELD_TYPE_TEXT },
			FIELD_CREDITCARD_EXPIRYDATE	: { "name" : "Expiry date",	"type" : FIELD_TYPE_TEXT },

			FIELD_PHONE_PHONENUMBER		: { "name" : "Phone number",	"type" : FIELD_TYPE_TEXT }
		}


	def get_field_data(self, field):
		"""Fetches data for a field type"""

		return self.fieldtype[field].copy()


	def get_list(self):
		"""Fetches a list of entry types"""

		list = self.entrytype.keys()
		list.sort()

		return list


	def get_data(self, type):
		"""Returns data for a type"""

		return self.entrytype[type]



class EntryStore(gtk.TreeStore):
	"""Base class for generic entry store"""

	def __init__(self, entrytypes):
		"""Sets up the entrystore"""

		gtk.TreeStore.__init__(self, gobject.TYPE_STRING, gobject.TYPE_STRING, gobject.TYPE_STRING, gobject.TYPE_STRING, gobject.TYPE_INT, gobject.TYPE_PYOBJECT)
		self.entrytypes = entrytypes


	def add_entry(self, parent, data = None):
		"""Adds an entry to the entry store"""

		iter = self.append(parent)
		self.update_entry(iter, data)
		return iter


	def export_entrystore(self, iter = None, entrystore = None, parent = None):
		"""Exports a branch to a new entrystore"""

		if entrystore == None:
			entrystore = EntryStore(EntryTypes())

		if iter != None:
			parent = entrystore.add_entry(parent, self.get_entry(iter))

		for i in range(self.iter_n_children(iter)):
			self.export_entrystore(self.iter_nth_child(iter, i), entrystore, parent)

		return entrystore


	def get_entry(self, iter):
		"""Fetches data for a tree entry"""

		if iter == None:
			return None
		else:
			return {
				"name"		: self.get_value(iter, ENTRYSTORE_COL_NAME),
				"type"		: self.get_value(iter, ENTRYSTORE_COL_TYPE),
				"icon"		: self.get_value(iter, ENTRYSTORE_COL_ICON),
				"description"	: self.get_value(iter, ENTRYSTORE_COL_DESC),
				"updated"	: self.get_value(iter, ENTRYSTORE_COL_UPDATED),
				"fields"	: self.get_value(iter, ENTRYSTORE_COL_FIELDS)
			}


	def get_entry_type(self, iter):
		"""Returns the type of an entry"""

		if iter == None:
			return None
		else:
			return self.get_value(iter, ENTRYSTORE_COL_TYPE)


	def has_contents(self):
		"""Checks if the entrystore has any contents"""

		if self.iter_n_children(None) > 0:
			return gtk.TRUE
		else:
			return gtk.FALSE


	def import_entrystore(self, entrystore, parent, importiter = None, baseiter = None):
		"""Imports entries from a different entrystore"""

		# add the first iter if specified, otherwise do all iters
		if importiter != None:
			childiter = self.add_entry(parent, entrystore.get_entry(importiter))

			if baseiter == None:
				baseiter = childiter
		else:
			childiter = parent

		# handle children
		for i in range(entrystore.iter_n_children(importiter)):
			baseiter = self.import_entrystore(entrystore, childiter, entrystore.iter_nth_child(importiter, i), baseiter)

		return baseiter


	def import_entrystore_after(self, entrystore, parent, sibling):
		"""Imports entries after an existing entry"""

		iter_root = entrystore.get_iter_root()
		iter = self.insert_entry_after(parent, sibling, entrystore.get_entry(iter_root))

		for i in range(entrystore.iter_n_children(iter_root)):
			self.import_entrystore(entrystore, iter, entrystore.iter_nth_child(iter_root, i))

		return iter


	def import_entrystore_before(self, entrystore, parent, sibling):
		"""Imports entries before an existing entry"""

		iter_root = entrystore.get_iter_root()
		iter = self.insert_entry_before(parent, sibling, entrystore.get_entry(iter_root))

		for i in range(entrystore.iter_n_children(iter_root)):
			self.import_entrystore(entrystore, iter, entrystore.iter_nth_child(iter_root, i))

		return iter


	def insert_entry_after(self, parent, sibling, data = None):
		"""Inserts an entry after a different entry"""

		iter = self.insert_after(parent, sibling)
		self.update_entry(iter, data)
		return iter


	def insert_entry_before(self, parent, sibling, data = None):
		"""Inserts an entry before a different entry"""

		iter = self.insert_before(parent, sibling)
		self.update_entry(iter, data)
		return iter


	def is_empty(self):
		"""Checks if the entrystore is empty"""

		if self.iter_n_children(None) == 0:
			return gtk.TRUE
		else:
			return gtk.FALSE


	def remove_entry(self, iter):
		"""Recursively removes an entry"""

		while self.iter_n_children(iter) > 0:
			self.remove_entry(self.iter_children(iter))

		self.remove(iter)


	def update_entry(self, iter, data):
		"""Updates data for an entry"""

		if data != None:
			self.set_value(iter, ENTRYSTORE_COL_NAME, data["name"])
			self.set_value(iter, ENTRYSTORE_COL_TYPE, data["type"])
			self.set_value(iter, ENTRYSTORE_COL_DESC, data["description"])
			self.set_value(iter, ENTRYSTORE_COL_UPDATED, data["updated"])
			self.set_value(iter, ENTRYSTORE_COL_FIELDS, data["fields"])
			self.set_value(iter, ENTRYSTORE_COL_ICON, self.entrytypes.get_data(data["type"])["icon"])



class DataStore(EntryStore):
	"""High-level subclass of entry store, for application use"""

	def __init__(self, entrytypes):
		"""Sets up the datastore"""

		EntryStore.__init__(self, entrytypes)
		self.changed = gtk.FALSE


	def add_entry(self, parent, data = None):
		"""Adds an entry to the data store"""

		# place inside parent if folder
		if self.get_entry_type(parent) in (ENTRY_FOLDER, None):
			iter = EntryStore.add_entry(self, parent, data)

		# place after "parent" if not folder
		else:
			iter = EntryStore.insert_entry_after(self, self.iter_parent(parent), parent, data)

		self.changed = gtk.TRUE

		return iter


	def clear(self):
		"""Removes all data from the datastore"""

		EntryStore.clear(self)
		self.changed = gtk.FALSE


	def export_xml(self, parent = None, level = 0):
		"""Generates xml based on data in the datastore"""

		xml = ""

		# process each child
		numchildren = self.iter_n_children(parent)
		for i in range(numchildren):

			# get child data
			iter = self.iter_nth_child(parent, i)
			data = self.get_entry(iter)

			tabs = "\t" * (level + 1)

			# generate entry
			xml = xml + "\n"
			xml = xml + tabs + "<entry type=\"" + escape_markup(data["type"]) + "\">\n"
			xml = xml + tabs + "	<name>" + escape_markup(data["name"]) + "</name>\n"
			xml = xml + tabs + "	<description>" + escape_markup(data["description"]) + "</description>\n"
			xml = xml + tabs + "	<updated>" + escape_markup(str(data["updated"])) + "</updated>\n"

			# generate fields
			for field in self.entrytypes.get_data(data["type"])["fields"]:
				xml = xml + tabs + "	<field id=\"" + escape_markup(field) + "\">" + escape_markup(data["fields"][field]) + "</field>\n"

			# handle any children
			if self.iter_n_children(iter) > 0:
				xml = xml + self.export_xml(iter, level + 1)

			xml = xml + tabs + "</entry>\n"

		# generate header and footer if at level 0
		if level == 0:
			header = "<?xml version=\"1.0\" ?>\n"
			header = header + "<revelationdata version=\"" + revelation.VERSION + "\">\n"
			footer = "</revelationdata>\n"

			xml = header + xml + footer

		return xml


	def get_entry(self, iter):
		"""Fetches data for a tree entry"""

		data = EntryStore.get_entry(self, iter)

		# always return the normal, closed folder icon for folders.
		# the open folder icon is only for tree-internal use.
		if data["icon"] == revelation.ui.STOCK_FOLDER_OPEN:
			data["icon"] = revelation.ui.STOCK_FOLDER

		return data


	def import_xml(self, xml):
		"""Imports data from xml"""

		try: doc = libxml2.parseDoc(xml)
		except TypeError:
			return gtk.FALSE

		# fetch and validate root
		root = doc.children
		if root.name != "revelationdata":
			return gtk.FALSE

		# process entries
		child = root.children
		while child is not None:
			if self.import_xml_node(child) == gtk.FALSE:
				return gtk.FALSE
			child = child.next

		return gtk.TRUE


	def import_xml_node(self, node, parent = None):
		"""Adds entries to the datastore recursively"""

		# skip text outside nodes
		if node.type == "text":
			return gtk.TRUE

		# skip unknown nodes
		if node.type != "element" or node.name != "entry":
			return gtk.FALSE

		# set up empty data structure
		data = {
			"name"		: "",
			"type"		: self.import_xml_get_attrs(node)["type"],
			"description"	: "",
			"updated"	: 0,
			"fields"	: {}
		}

		# add empty entry, iter needed for any children
		iter = self.add_entry(parent)

		# get elements
		child = node.children
		while child is not None:

			if child.type == "element":

				if child.name == "name":
					data["name"] = child.content

				elif child.name == "description":
					data["description"] = child.content

				elif child.name == "updated":
					data["updated"] = int(child.content)

				elif child.name == "field":
					field = self.import_xml_get_attrs(child)["id"]
					data["fields"][field] = child.content

				elif child.name == "entry":
					self.import_xml_node(child, iter)

				else:
					return gtk.FALSE

			child = child.next

		# set up missing fields according to entry type
		for field in self.entrytypes.get_data(data["type"])["fields"]:
			if data["fields"].has_key(field) == False:
				data["fields"][field] = ""

		# update entry with actual data
		self.update_entry(iter, data)

		return gtk.TRUE


	def import_xml_get_attrs(self, node):
		"""Returns a dict of attributes for a node"""

		attrs = {}

		attr = node.properties
		while attr is not None:
			attrs[attr.name] = attr.content
			attr = attr.next

		return attrs


	def remove_entry(self, iter):
		"""Removes an entry and all children"""

		parent = self.iter_parent(iter)

		# remove entry
		EntryStore.remove_entry(self, iter)

		# collapse parent if empty
		if self.iter_n_children(parent) == 0:
			self.set_folder_state(parent, gtk.FALSE)

		self.changed = gtk.TRUE


	def set_folder_state(self, iter, open):
		"""Sets the open/closed state of a folder, changing its icon"""

		if iter == None or self.get_entry_type(iter) != ENTRY_FOLDER:
			return

		if open == gtk.TRUE:
			icon = revelation.ui.STOCK_FOLDER_OPEN
		else:
			icon = revelation.ui.STOCK_FOLDER

		self.set_value(iter, ENTRYSTORE_COL_ICON, icon)


	def update_entry(self, iter, data):
		"""Updates data for an entry"""

		curicon = self.get_value(iter, ENTRYSTORE_COL_ICON)

		EntryStore.update_entry(self, iter, data)

		# keep icon if current is folder-open and the type still is folder
		if curicon == revelation.ui.STOCK_FOLDER_OPEN and data["type"] == ENTRY_FOLDER:
			self.set_value(iter, ENTRYSTORE_COL_ICON, revelation.ui.STOCK_FOLDER_OPEN)

		self.changed = gtk.TRUE


class EntryClipboard(EntryStore):
	"""An entry clipboard"""

	def __init__(self):

		EntryStore.__init__(self, EntryTypes())


	def copy(self, datastore, iter):
		"""Copies data from the datastore"""

		if iter == None:
			return

		self.clear()
		self.import_entrystore(datastore, None, iter)
		self.emit("copy")


	def cut(self, datastore, iter):
		"""Cuts data from the datastore"""

		if iter != None:
			self.copy(datastore, iter)
			datastore.remove_entry(iter)
			self.emit("cut")


	def paste(self, datastore, parent, sibling = None):
		"""Pastes data into the datastore"""

		if self.is_empty() == gtk.TRUE:
			return

		if sibling == None:
			iter = datastore.import_entrystore(self, parent)
		else:
			iter = datastore.import_entrystore_after(self, parent, sibling)

		self.emit("paste")

		return iter


gobject.signal_new("copy", EntryClipboard, gobject.SIGNAL_ACTION, gobject.TYPE_BOOLEAN, ())
gobject.signal_new("cut", EntryClipboard, gobject.SIGNAL_ACTION, gobject.TYPE_BOOLEAN, ())
gobject.signal_new("paste", EntryClipboard, gobject.SIGNAL_ACTION, gobject.TYPE_BOOLEAN, ())



class UndoQueue(gobject.GObject):

	def __init__(self):

		gobject.GObject.__init__(self)

		self.queue = []
		self.actionptr = 0


	def add_action(self, name, action, data):

		# remove any items later in the queue
		del self.queue[self.actionptr:]

		self.queue.append({ "name" : name, "action" : action, "data" : data })
		self.actionptr = len(self.queue)

		self.emit("can-undo", self.can_undo())
		self.emit("can-redo", self.can_undo(REDO))


	def can_undo(self, method = UNDO):
		"""Checks if the undo queue allows an undo action"""

		if method == UNDO:
			return self.actionptr > 0
		else:
			return self.actionptr < len(self.queue)


	def clear(self):
		"""Clears the undo queue"""

		self.queue = []
		self.actionptr = 0

		self.emit("can-undo", gtk.FALSE)
		self.emit("can-redo", gtk.FALSE)


	def get_data(self, method = UNDO):
		"""Returns the current undo data"""

		if method == UNDO:
			ptr = self.actionptr - 1
		else:
			ptr = self.actionptr

		if self.can_undo(method) == gtk.TRUE:
			return self.queue[ptr]
		else:
			return None


	def redo(self):
		"""Redoes the next action"""

		if self.can_undo(REDO) == gtk.TRUE:
			self.emit("redo", self.get_data(REDO))
			self.actionptr = self.actionptr + 1

		self.emit("can-undo", self.can_undo())
		self.emit("can-redo", self.can_undo(REDO))


	def undo(self):
		"""Undoes the previous action"""

		if self.can_undo() == gtk.TRUE:
			self.emit("undo", self.get_data())
			self.actionptr = self.actionptr - 1

		self.emit("can-undo", self.can_undo())
		self.emit("can-redo", self.can_undo(REDO))


gobject.signal_new("undo", UndoQueue, gobject.SIGNAL_ACTION, gobject.TYPE_BOOLEAN, (gobject.TYPE_PYOBJECT, ))
gobject.signal_new("redo", UndoQueue, gobject.SIGNAL_ACTION, gobject.TYPE_BOOLEAN, (gobject.TYPE_PYOBJECT, ))
gobject.signal_new("can-undo", UndoQueue, gobject.SIGNAL_ACTION, gobject.TYPE_BOOLEAN, (gobject.TYPE_BOOLEAN, ))
gobject.signal_new("can-redo", UndoQueue, gobject.SIGNAL_ACTION, gobject.TYPE_BOOLEAN, (gobject.TYPE_BOOLEAN, ))


def decrypt(data, password):
	"""decrypts data"""

	padlen = ENCRYPT_KEYSIZE - (len(password) % ENCRYPT_KEYSIZE)
	AES.key_size = ENCRYPT_KEYSIZE
	cipher = AES.new(password + (chr(0) * padlen))

	if len(data) % cipher.block_size != 0:
		return ""

	decrypted = cipher.decrypt(data)

	return decrypted


def encrypt(data, password):
	"""encrypts data"""

	padlen = ENCRYPT_KEYSIZE - (len(password) % ENCRYPT_KEYSIZE)
	AES.key_size = ENCRYPT_KEYSIZE
	cipher = AES.new(password + (chr(0) * padlen))

	padlen = cipher.block_size - (len(data) % cipher.block_size)
	if padlen == cipher.block_size:
		padlen = 0
	encrypted = cipher.encrypt(data + (" " * padlen))

	return encrypted


def escape_markup(string):
	"""Escapes special characters in a string"""

	string = string.replace("&", "&amp;")
	string = string.replace("<", "&lt;")
	string = string.replace(">", "&gt;")

	return string


def file_read(file):
	"""reads data from a file"""

	try: fp = open(file, "rb", 0)
	except IOError:
		return None

	data = fp.read()
	fp.close()

	return data


def file_write(file, data):
	"""writes data to a file"""

	try: fp = open(file, "wb", 0)
	except IOError:
		return gtk.FALSE

	fp.write(data)
	fp.flush()
	fp.close()

	return gtk.TRUE

